# -*- coding: utf-8 -*-

import asyncio
import logging
import weakref
from collections import namedtuple
from functools import partial
from typing import Coroutine, MutableMapping, Awaitable

from patchwork.core import Component
from patchwork.core.utils import Flag, cached_property

Job = namedtuple('Job', ('task', 'coroutine'))


class ProcessingUnit(Component):

    class Config(Component.Config):
        # number of tasks which can be queued for waiting until processing unit is ready
        queue_size: int = 1

        # number of tasks which can be run concurrently on the processing unit
        concurrency: int = 1

    _processing_loop: asyncio.Future
    _in_queue: asyncio.Queue
    _available: asyncio.Event
    _busy: Flag

    def __init__(self, *, parent, **options):
        self._jobs = set()
        super().__init__(parent=parent, **options)

    @property
    def event_loop(self) -> asyncio.AbstractEventLoop:
        return asyncio.get_running_loop()

    @cached_property
    def logger(self) -> logging.Logger:
        return logging.getLogger('patchwork.worker.unit')

    @property
    def busy(self) -> Flag:
        """
        Returns awaitable flag which is True when unit is busy (which means there is at least one task in progress).
        :return:
        """
        return self._busy

    @property
    def executor(self):
        return self.parent

    async def _start(self):
        assert self.settings.queue_size > 0, "Queue length cannot be less than 1"
        assert self.settings.concurrency > 0, "Concurrency cannot be less than 1"

        self._in_queue = asyncio.Queue(maxsize=self.settings.queue_size)
        self._available = asyncio.Event()
        self._available.set()
        self._busy = Flag(False)

    async def run(self):
        await super().run()
        self._processing_loop = asyncio.create_task(self._loop())
        self._processing_loop.add_done_callback(self._on_loop_completed)

    def _on_loop_completed(self, fut: asyncio.Future):
        if fut.cancelled():
            return

        exc = fut.exception()
        if exc is not None:
            self.logger.error(f"Processing unit crashed with exception: {exc.__class__.__name__}({exc})")

        # unit loop is not working, this unit is no longer usable
        stop = asyncio.create_task(self.terminate())

        def _cblk(f):
            if f.cancelled() or f.exception():
                # stop failed, force state flag to false so parent component can react
                self.state.set(False)
                self.logger.fatal(f"Can't stop crashed processing unit. Termination forced")

        stop.add_done_callback(_cblk)

    async def _loop(self):
        while True:
            try:
                # wait until unit is available
                await self._available.wait()

                # unit is ready, wait for task
                # fut is a Future which represents task to run, this future has been
                # returned to the callee who scheduled the coro to run
                job: Job = await self._in_queue.get()

                if job.task.cancelled():
                    # job has been cancelled in meanwhile (between scheduling and now)
                    continue

                # create async background task from given coroutine
                runner = self.event_loop.create_task(self._run_job(job.coroutine))
                self._jobs.add(job)
                self.logger.debug(f"Task `{job.coroutine.cr_code.co_name}` scheduled")

                runner_callback = partial(self._on_job_done, job=job)
                runner.add_done_callback(runner_callback)

                # when task is cancelled, remove listener which transfers result from runner to the task and
                # cancel the runner future
                job.task.add_done_callback(lambda f: runner.remove_done_callback(runner_callback) and runner.cancel())

                if len(self._jobs) == self.settings.concurrency:
                    self._available.clear()

                self._busy.set(True)

            except asyncio.CancelledError:
                return

    def _on_job_done(self, fut: asyncio.Future, job: Job):
        """
        Scheduled coroutine future callback. The main job here is to pass fut state into task.
        :param fut: Future which represents scheduled coroutine (code)
        :param job:
        :return:
        """
        if fut.cancelled():
            job.task.cancel()

        exc = fut.exception()
        if exc is not None:
            job.task.set_exception(exc)
        else:
            job.task.set_result(fut.result())

        self._jobs.remove(job)
        self._in_queue.task_done()
        if len(self._jobs) < self.settings.concurrency:
            self._available.set()

        if not self._jobs:
            self._busy.set(False)

    async def submit(self, coro: Coroutine):
        """
        Submits given coroutine to run on the worker, return a coroutine which represents submitted task.
        Result of returned coro is the same as result of submitted coro, canceling it cancels submitted task,
        however returned coro may raise more exceptions than submitted one.
        :param coro:
        :return:
        """
        fut = self.event_loop.create_future()
        await self._in_queue.put(Job(task=fut, coroutine=coro))
        return fut

    async def _stop(self):
        """
        Terminate worker.
        :return:
        """
        self.logger.debug(f"{self.__class__.__name__} received termination request, "
                          f"there is {len(self._jobs)} active tasks")

        # cancel processing unit loop
        if not self._processing_loop.done():
            self._processing_loop.cancel()

        # request cancelling of all running tasks
        for job in self._jobs:
            job.task.cancel()

        if self._jobs:
            self.logger.debug(f"Procesing unit has {len(self._jobs)} active jobs. Waiting for cancellation...")
            # wait until tasks are completed
            await asyncio.wait([job.task for job in self._jobs])

    def _run_job(self, coro: Coroutine) -> Awaitable:
        """
        Schedules given coroutine to run and returns awaitable object which represents
        coro execution.
        :param coro:
        :return:
        """
        raise NotImplementedError()

    def get_state(self) -> MutableMapping:
        """
        Returns current worker state
        :return:
        """
        return {
            'tasks': len(self._jobs)
        }

    def __repr__(self):
        res = super().__repr__()
        return f"<{res[1:-1]} [{'available' if self._available.is_set() else 'full'}, " \
               f"qsize={self.settings.queue_size}, concurrency={self.settings.concurrency}, tasks={len(self._jobs)}]>"



# -*- coding: utf-8 -*-

from typing import Coroutine, Awaitable

from .base import ProcessingUnit


class InlineUnit(ProcessingUnit):
    """
    Simple processing unit which runs tasks one the same event loop
    as processing unit.
    """

    def _run_job(self, coro: Coroutine) -> Awaitable:
        return coro




# -*- coding: utf-8 -*-
from typing import Any, Mapping

from patchwork import core
from patchwork.core.utils import cached_property


class Context:
    def __init__(self, worker: 'PatchworkWorker', task: core.Task, receiver: core.AsyncSubscriber):
        self.worker = worker
        self.task = task
        self.receiver = receiver

    @cached_property
    def publishers(self) -> Mapping[str, core.AsyncPublisher]:
        return self.worker.publishers

    @property
    def payload(self):
        return self.task.payload


class Dependency:

    def resolve(self, ctx):
        raise NotImplementedError()


class Depends(Dependency):

    def __init__(self, call):
        self.call = call

    def resolve(self, ctx):
        return self.call()


class GetPublisher(Dependency):

    def resolve(self, ctx: Context):
        return ctx.worker.get_publisher()


class GetPayload(Dependency):
    def __init__(self, schema: Any):
        self.schema = schema

    def resolve(self, ctx):
        obj = self.schema()
        ctx.payload.Unpack(obj)
        return obj


class GetTask(Dependency):

    def resolve(self, ctx: Context):
        return ctx.task



# -*- coding: utf-8 -*-
from contextlib import contextmanager

import logging

import pytest


@pytest.fixture
def suppress_logging():

    @contextmanager
    def _suppresor(max_level=logging.WARNING):
        logging.disable(max_level)
        yield
        logging.disable(logging.NOTSET)
    return _suppresor



# -*- coding: utf-8 -*-
import fnmatch
import weakref

import asyncio

import inspect
from google.protobuf.reflection import GeneratedProtocolMessageType
from pydantic.typing import ForwardRef, evaluate_forwardref

from typing import Callable, Any, Dict, List

from patchwork.core import Task, AsyncSubscriber
from patchwork.node.core import dependencies


def get_typed_annotation(param: inspect.Parameter, globalns: Dict[str, Any]) -> Any:
    annotation = param.annotation
    if isinstance(annotation, str):
        annotation = ForwardRef(annotation)
        annotation = evaluate_forwardref(annotation, globalns, globalns)
    return annotation


def get_typed_signature(call: Callable[..., Any]) -> inspect.Signature:
    signature = inspect.signature(call)
    globalns = getattr(call, "__globals__", {})
    typed_params = [
        inspect.Parameter(
            name=param.name,
            kind=param.kind,
            default=param.default,
            annotation=get_typed_annotation(param, globalns),
        )
        for param in signature.parameters.values()
    ]
    typed_signature = inspect.Signature(typed_params)
    return typed_signature


class Handler:

    def __init__(self, call):
        self.call = call
        sig = get_typed_signature(call)
        self.kwargs = {}

        for param_name, param in sig.parameters.items():
            if isinstance(param.default, dependencies.Dependency):
                self.kwargs[param_name] = param.default
                continue
            if isinstance(param.annotation, Task):
                self.kwargs[param_name] = dependencies.GetTask()
                continue
            if isinstance(param.annotation, GeneratedProtocolMessageType):
                self.kwargs[param_name] = dependencies.GetPayload(param.annotation)
                continue

            raise NotImplementedError(f"{param_name} is not supported")

    def __call__(self, ctx: dependencies.Context):
        kwargs = {k: v.resolve(ctx) for k, v in self.kwargs.items()}
        return self.call(**kwargs)


class Route:

    def __init__(self, handler: Handler, path: str):
        self.handler = handler
        self.path = path

    def match(self, path):
        return fnmatch.fnmatch(path, self.path)

    def __call__(self, ctx: dependencies.Context):
        return self.handler(ctx)


class TaskRouter:

    def __init__(self):
        self.routes: List[Route] = []
        self._worker = None

    def bind(self, worker):
        self._worker = weakref.ref(worker)

    def on(self, *task_type: str):
        def decorator(fn):
            assert asyncio.iscoroutinefunction(fn), "Handler must be an async function"

            handler = Handler(fn)

            for tt in task_type:
                self.routes.append(Route(handler, tt))

            return fn
        return decorator

    def handle(self, task: Task, receiver: AsyncSubscriber):
        for route in self.routes:
            if route.match(task.task_type):
                ctx = dependencies.Context(self._worker(), task, receiver)
                return route(ctx)



# -*- coding: utf-8 -*-
import mock
import pytest

from patchwork.node.core.exceptions import TaskControlException
from patchwork.node.executor import BaseExecutor
from patchwork.node.executor.base import default_task_exception_handler


@pytest.mark.asyncio
async def test_no_custom_regular_exception():
    executor = BaseExecutor(parent=None)

    handlers = executor._get_exception_handler(Exception())
    assert list(handlers) == [default_task_exception_handler]


@pytest.mark.asyncio
async def test_custom_same_level_handler_regular_exception():
    executor = BaseExecutor(parent=None)

    async def handler():
        pass

    executor.add_exception_handler(Exception, handler)

    handlers = executor._get_exception_handler(Exception())
    assert list(handlers) == [handler, default_task_exception_handler]


@pytest.mark.asyncio
async def test_subclassing_order():

    executor = BaseExecutor(parent=None)

    class CustomException(Exception):
        pass

    class SubCustomException(CustomException):
        pass

    async def custom_exc_handler():
        pass

    async def generic_handler():
        pass

    executor.add_exception_handler(CustomException, custom_exc_handler)
    executor.add_exception_handler(Exception, generic_handler)

    handlers = executor._get_exception_handler(SubCustomException())

    # custom_exc_handler should be called before any handler registered for Exception, because
    # it's more specific - tied to CustomException which is better matching raised exception
    assert list(handlers) == [custom_exc_handler, generic_handler, default_task_exception_handler]



# -*- coding: utf-8 -*-
from patchwork.core import Task

from patchwork.node.core import TaskRouter
from patchwork.node.core.exceptions import TaskRetry, TaskFatal, TaskDrop, TaskControlException


class CustomException(Exception):
    pass


test_router = TaskRouter()


class UnknownTCE(TaskControlException):
    pass


@test_router.on('noop')
async def processor_noop():
    return None


@test_router.on('exception')
async def processor_exception():
    raise CustomException()


@test_router.on('retry')
async def processor_retry():
    raise TaskRetry()


@test_router.on('fatal')
async def processor_fatal():
    raise TaskFatal()


@test_router.on('drop')
async def processor_drop():
    raise TaskDrop()


@test_router.on('unknown_tce')
async def processor_drop():
    raise UnknownTCE()



# -*- coding: utf-8 -*-
import asyncio

import pytest as pytest

from patchwork.core import Task
from patchwork.node import PatchworkWorker
from patchwork.node.core.router import TaskRouter


@pytest.fixture
async def worker():
    worker = PatchworkWorker()
    await worker.run()
    yield worker

    await worker.terminate()


@pytest.mark.asyncio
async def test_simple_use(worker):

    caught_tasks = []

    router = TaskRouter()

    @router.on('test-msg')
    async def processor(task: Task):
        caught_tasks.append(task)

    worker.include_router(router)
    await worker.get_publisher().send_task(Task(task_type='test-msg'))

    # wait until executor processing unit becomes busy which means given message has been consumed
    await asyncio.wait_for(worker.executor.unit.busy.wait(), timeout=1)

    assert len(caught_tasks) == 1



# -*- coding: utf-8 -*-

from contextlib import asynccontextmanager

from typing import Type, Mapping

from patchwork.core.testutils.cases import AsyncioTestCase
from patchwork.node.core import Module
from patchwork.node.testutils.mocks import WorkerMock


class ModuleTestCase(AsyncioTestCase):

    module_class: Type[Module]
    worker: WorkerMock

    def get_module_settings(self) -> Mapping:
        return {}

    def setUp(self) -> None:
        super().setUp()
        self._worker = WorkerMock()

        self.module = self.module_class(self._worker, **self.get_module_settings())

    def tearDown(self) -> None:
        del self.module
        del self._worker

    async def _test_module_run(self):

        self.assertTrue(await self.module.run(), msg="Module run() method should return True "
                                                     "when successfully started")

        self.assertTrue(self.module.is_running, msg="Module should have is_running property set to True after run(). "
                                                    "Is run()/start() methods return when module is really started?")

        self.assertTrue(await self.module.terminate(), msg="Module terminate() should return True "
                                                           "when successfully stopped")
        self.assertFalse(self.module.is_running, msg="Module should have is_running property set to False after"
                                                     "terminate(). Is terminate()/stop() methods return when module"
                                                     "is really stopped?")

    def test_00_module_lifetime(self):
        # name is prefixed with '00' to run it as the first one (in most cases) so you can run test with failfast
        # and if module startup and shutdown is not passing then running all other tests make no sense and won't
        # be executed
        self.loop.run_until_complete(self._test_module_run())

    @asynccontextmanager
    async def running_module(self):
        if not await self.module.run():
            raise RuntimeError("Module didn't start, can't run test")

        try:
            yield self.module
        finally:
            await self.module.terminate()



# -*- coding: utf-8 -*-

import weakref
from asyncio import AbstractEventLoop
from dataclasses import field

import asyncio
import multiprocessing
import random
from asyncio.futures import Future
from functools import partial
from threading import Thread
from typing import Coroutine, List, Union, Tuple, Type, Awaitable

from patchwork.core.typing import ClassPath
from .base import ProcessingUnit

DEFAULT_CONFIG = {
    'workers': multiprocessing.cpu_count(),
    'assignor': 'collage_worker.executor.slots.threads:RandomThreadAssignor'
}


class _Thread:
    """
    Represents single worker thread. Processing unit spawns multiple _Thread's, each with it's own event loop
    and assigns tasks to them.
    """

    _thread: Thread
    _future: Future

    def __init__(self, name):
        self._name = name
        self._thread_started = asyncio.Event()
        self._is_stopping = False

    def _ready(self, loop: AbstractEventLoop):
        self._loop_ref = weakref.ref(loop, self._ref_gone_callback)
        loop.call_soon_threadsafe(self._thread_started.set)

    def _ref_gone_callback(self):
        if self._is_stopping:
            # if ref gone due to thread stop, consider this as cancel
            self._future.cancel()
        else:
            # otherwise thread just completed, don't know why but thread should not complete, never
            self._future.set_exception(RuntimeError("_Thread unit completed"))

    @property
    def loop(self):
        loop = self._loop_ref()
        if loop is None or not loop.is_running():
            self._future.set_exception(RuntimeError("_Thread unit dead and event loop cannot be accessed"))
            return None
        return loop

    @staticmethod
    def _main(ready_callback):
        loop = asyncio.new_event_loop()
        ready_callback(loop)
        loop.run_forever()

    @staticmethod
    async def _worker_stop_job(loop, stop_timeout):
        # Careful, this code is running on thread event loop!

        self_task = asyncio.tasks.Task.current_task(loop=loop)

        tasks = [t for t in asyncio.Task.all_tasks(loop=loop) if not t.done() and t is not self_task]
        for task in tasks:
            task.cancel()

        done, pending = await asyncio.wait(tasks, timeout=stop_timeout)

        loop.terminate()
        return pending

    async def start(self) -> Future:
        """
        Starts worker and waits until started. Returns future which is completed when underlying worker
        thread is completed. If worker is stopped future is marked as cancelled. If worker die unexpectedly
        future has RuntimeError exception raised.
        :return:
        """
        self._future = Future()
        self._thread = Thread(target=self._main, name=self._name, daemon=True, kwargs={
            'ready_callback': self._ready
        })
        await self._thread_started.wait()

        return self._future

    async def stop(self, stop_timeout):
        """
        Stop worker and wait for all internal jobs to complete for stop_timeout period.
        Returns list of jobs which were not completed in given timeout.
        :return:
        """
        self._is_stopping = True
        pending = await self.submit(self._worker_stop_job(self.loop, stop_timeout))

        await self._future

        return pending

    def submit(self, coro: Coroutine) -> Future:
        """
        Submits given coroutine to run on worker thread event loop.
        :param coro:
        :return:
        """
        if self._is_stopping:
            raise RuntimeError("Can't schedule coroutine, worker is stopping")

        return asyncio.run_coroutine_threadsafe(coro, self.loop)

    def __bool__(self):
        """
        Returns if worker is alive and operational.
        :return:
        """
        return not self._future.done()

    def tasks(self):
        """
        Returns list of tasks which are active on the worker
        :return:
        """
        return tuple(t for t in asyncio.Task.all_tasks(loop=self.loop) if not t.done())


class AbstractAssignor:

    def __call__(self, workers: Tuple[_Thread]) -> _Thread:
        raise NotImplementedError()


class RandomAssignor(AbstractAssignor):
    """
    Randomly assigns jobs to workers
    """

    def __call__(self, workers: Tuple[_Thread]) -> _Thread:
        return random.choice(workers)


class LessTasksAssignor(AbstractAssignor):
    """
    Assigns jobs to worker which has lowest number of active tasks.
    Note: this does not mean that this worker is the most free one, there could be one heavy task
    assigned to the worker and many mostly waiting tasks assigned to other workers
    """

    def __call__(self, workers: Tuple[_Thread]) -> _Thread:
        min_tasks = None
        worker = None

        for w in workers:
            tasks = len(w.tasks())
            if min_tasks is None or tasks < min_tasks:
                min_tasks = tasks
                worker = w

        return worker


class ThreadPoolUnit(ProcessingUnit):
    """
    Run jobs in separate threads async loops to avoid blocking main thread loop
    when jobs are CPU intensive or blocking. Tasks are distributed among threads
    by assignor.
    """

    class Config(ProcessingUnit.Config):
        pass
        # use given assignor to assign tasks to thread workers
        # assignor: ClassPath = 'collage_worker.executor.workers.threads.RandomAssignor'

        # number of thread workers to spawn, by default number of CPUs
        # threads: int = field(default_factory=multiprocessing.cpu_count)

        # assignor options, depends on actual assignor, all options will be passed as __init__ arguments
        # to assignor class
        # assignor_options = field(default_factory=dict)

    _threads: List[Union[None, _Thread]]
    _assignor: AbstractAssignor

    def __repr__(self):
        res = super().__repr__()
        return f"<{res[1:-2]}, threads={self.settings.threads}, assignor={self.settings.assignor.__class__.__name__}]>"

    async def _run_job(self, coro: Coroutine) -> Awaitable:
        workers: Tuple[_Thread] = tuple(w for w in self._threads if w is not None)

        choosen_worker: _Thread = self._assignor(workers)
        return choosen_worker.submit(coro)

    async def _start(self) -> bool:
        """
        Setup worker
        :return:
        """

        self._threads: List[Union[None, _Thread]] = [None] * self.settings.threads
        try:
            assignor_cls: Type[AbstractAssignor] = self.settings.assignor
        except Exception as e:
            self.logger.error(
                f"Can't load assignor class from '{self.settings.assignor}'\nDetails:\n{type(e)}: {e}",
                exc_info=True,
                extra={
                    'assignor': self.settings.assignor
                }
            )
            return False

        try:
            self._assignor = assignor_cls(**self.settings.assignor_options)
        except Exception as e:
            self.logger.error(
                f"Can't instantiate assignor '{assignor_cls.__name__}'\nDetails:\n{type(e)}: {e}",
                exc_info=True,
                extra={
                    'assignor': self.settings.assignor
                }
            )
            return False

        self.logger.info(f"Setting up {self.settings.threads} thread workers")

        for idx in range(self.settings.threads):
            await self._spawn(idx)

        return True

    async def _spawn(self, index):
        assert self._threads[index] is None

        self._threads[index] = _Thread(name=f'unit-{index}')

        # start returns a future which represents worker thread, when done
        # it means that thread is completed
        fut = await self._threads[index].start()
        fut.add_done_callback(partial(self._on_worker_die, index=index))

    def _on_worker_die(self, fut: Future, index):
        if fut.cancelled():
            # worker has been cancelled, it's intentional "die"
            return

        exc = fut.exception()
        if exc:
            self.logger.error(f"{self.__class__.__name__} died: {exc}")
        else:
            self.logger.info(f"{self.__class__.__name__} completed with result '{fut.result()}'")

        self._threads[index] = None

        # schedule worker re-spawn
        asyncio.create_task(asyncio.ensure_future(self._spawn(index)))



# -*- coding: utf-8 -*-

import weakref
from asyncio import AbstractEventLoop, Future
from multiprocessing.managers import BaseManager

import asyncio
import importlib_metadata
import logging
import logging.config
import multiprocessing
import os
import psutil
import signal
import sys
import threading
import warnings
from datetime import datetime
from functools import partial
from humanize import naturalsize, naturaldelta
from pydantic import BaseModel
from time import time
from typing import Mapping, MutableMapping, Iterable, Dict, Type, Optional, Literal

from patchwork.core import AsyncPublisher, AsyncSubscriber, Component
from patchwork.core.client.local import AsyncLocalPublisher, AsyncLocalSubscriber
from patchwork.core.config.base import ComponentConfig, PublisherConfig, SubscriberConfig
from patchwork.core.typing import LoggingConfig
from patchwork.core.utils import cached_property
from patchwork.node.core import Module
from patchwork.node.core.router import TaskRouter
from patchwork.node.executor import BaseExecutor
from patchwork.node.manager.base import Manager
from patchwork.node.manager.http import HttpManager

if sys.version_info < (3, 9):
    from functools import lru_cache
    cache = lru_cache(maxsize=None)
else:
    from functools import cache


class WorkerSettings(BaseModel):

    debug: bool = False
    enabled: bool = True

    logging: LoggingConfig = None

    # executor unit
    executor: ComponentConfig[Type[BaseExecutor]] = ComponentConfig(BaseExecutor)

    # manager unit
    manager: ComponentConfig[Type[BaseManager]] = ComponentConfig(HttpManager, {
        'listen_on': '127.0.0.1:4444'
    })

    # optional list of additional modules to load
    modules: Dict[str, ComponentConfig] = {}

    publisher: Optional[PublisherConfig] = PublisherConfig(AsyncLocalPublisher)

    # clients for queues
    subscriber: SubscriberConfig = SubscriberConfig(AsyncLocalSubscriber)


class PatchworkWorker(Module):

    # worker settings dataclass
    settings: WorkerSettings

    _event_loop: AbstractEventLoop
    monitor_task: Future

    def __init__(self, settings_class: type = WorkerSettings, settings: Mapping = None, **kwargs):
        super().__init__(**kwargs)
        self._settings_schema = settings_class
        self._load_settings(settings)

    @cached_property
    def logger(self) -> logging.Logger:
        return logging.getLogger('patchwork.worker')

    @property
    def event_loop(self):
        return self._event_loop

    @property
    def process_name(self):
        return f'#{os.getpid()}: {multiprocessing.current_process().name}'

    @cached_property
    def manager(self) -> Manager:
        """
        Returns worker manager instance
        :return:
        """
        return self.settings.manager.instantiate(parent=self)

    @cached_property
    def executor(self) -> BaseExecutor:
        """
        Returns worker executor instance
        :return:
        """
        return self.settings.executor.instantiate(parent=self)

    @cached_property
    def modules(self):
        return {k: v.instantiate(parent=self) for k, v in self.settings.modules.items()}

    @property
    def debug(self):
        return self.settings.debug

    def middleware(self, flow: Literal['task'] = 'task'):

        if flow == 'task':
            def wrapper(fn):
                assert asyncio.iscoroutinefunction(fn)
                self.executor.add_middleware(fn)
                return fn
        else:
            raise NotImplementedError()

        return wrapper

    def exception_handler(self, exception_class: type):
        def wrapper(fn):
            assert asyncio.iscoroutinefunction(fn)
            self.executor.add_exception_handler(exception_class, fn)
            return fn

        return wrapper

    def _load_settings(self, settings: Mapping = None):
        """
        Load settings from worker settings JSON file and then updates values
        using environment variables if JSON file has no `locked_settings` flag enabled.
        :return:
        """
        try:
            self.settings = self._settings_schema(**(settings or {}))
        except Exception as e:
            self.logger.error(f"Unable to load settings: {e}")
            return False

        if self.settings.debug:
            warnings.warn("Debug mode enabled")
        return True

    def get_module(self, name: str) -> Module:
        """
        Returns module instance by given name
        :param name:
        :return:
        """
        return self.modules[name]

    @cache
    def get_publisher(self) -> AsyncPublisher:
        if self.settings.publisher is None:
            raise RuntimeError("can't use publisher as it's not configured")

        return self.settings.publisher.instantiate(self)

    @cache
    def get_subscriber(self) -> AsyncSubscriber:
        return self.settings.subscriber.instantiate(self)

    def main(self):
        """
        Runs the worker. This method is blocking until worker receive TERM signal.
        :return:
        """
        ev_loop = self._get_event_loop()

        # put executor start at event loop
        start_task = ev_loop.create_task(self.run())
        start_task.add_done_callback(self._on_started)

        # noinspection PyBroadException
        try:
            self.logger.debug("Starting worker event loop")
            ev_loop.run_forever()
            self.logger.debug("Worker event loop completed")
        except Exception:
            self.logger.fatal("Worker execution ends with unexpected exception", exc_info=True)

    def _on_started(self, fut: Future):
        exc = fut.exception()
        if exc:
            self.logger.error(f"Worker cannot be started due to exception: {type(exc).__name__}({exc})")

            # stop event loop if executor is not running
            # TODO: some cleanups are necessary?
            self._event_loop.stop()

            if getattr(self.settings, "debug", False):
                raise exc

            return

        if fut.result() is False:
            self.logger.error("Worker didn't start. See logs above for more details")
            self.terminate_worker()
            return

        self.logger.info("Node is up and running")

        # noinspection PyBroadException
        try:
            node_version = importlib_metadata.version('patchwork.node')
        except importlib_metadata.PackageNotFoundError:
            node_version = 'development'

        if self.settings.debug:
            print("\nPatchwork Node is up and running.")
            print(f"Version: {node_version}")
            print("Send TERM or INT signal to terminate")
            print("DEBUG MODE ENABLED, DO NOT RUN IN DEBUG ON PRODUCTION")

    async def _start(self):

        assert threading.current_thread() is threading.main_thread(), "Worker must start in MainThread"
        self._event_loop = asyncio.get_running_loop()

        self._setup_signals()

        self._setup_logger()
        self.logger.debug("Worker initialized")

        # first starts manager
        self.logger.debug("Starting manager")
        await self.manager.run()
        self.logger.info("Manager started")

        # then start additional modules, all at once, if modules depends on each other should await
        # internally. Use get_module() worker method to get module instance and await on state flag
        # Be careful of deadlocks

        start_jobs = [mod.run() for mod in self.modules.values()]
        if start_jobs:
            self.logger.debug("Starting additional modules")
            await asyncio.wait(start_jobs)
            self.logger.info("All additional modules started")
        else:
            self.logger.debug("No additional modules to start")

        # start clients
        if self.settings.publisher is not None:
            await self.get_publisher().run()
            self.logger.info("Publisher started")
        else:
            self.logger.debug("No publisher configured")

        await self.get_subscriber().run()
        self.logger.info("Subscriber started")

        # end executor at the end when everything is ready
        self.logger.debug("Starting executor")
        await self.executor.run()
        self.logger.info("Executor started")

        self.monitor_task = asyncio.create_task(self._monitor_submodules(
            (self.manager, self.executor) + tuple(self.modules.values())
        ))
        self.monitor_task.add_done_callback(self._on_monitor_stop)

    async def _monitor_submodules(self, submodules: Iterable[Component]):
        flags = {asyncio.ensure_future(s_mod.state.wait()): weakref.ref(s_mod) for s_mod in submodules}
        while True:
            try:
                done, pending = await asyncio.wait(flags.keys(), return_when=asyncio.FIRST_COMPLETED)

                if len(done) != 1:
                    self.logger.error(f"Monitor got invalid number of done flags, expected: 1, got: {len(flags)}")
                    continue

                done = done.pop()
                s_mod_ref = flags.pop(done)

                # dereference
                s_mod = s_mod_ref()

                if s_mod is None:
                    self.logger.warning(f"Stopped module disappeared")
                    continue

                self.logger.warning(f"{s_mod} terminated unexpectedly")
                recovered = await s_mod.recover()
            except asyncio.CancelledError:
                # cancel all active waiters
                for fut in flags.keys():
                    fut.cancel()

                raise
            else:
                if recovered:
                    # recovered!
                    flags[asyncio.ensure_future(s_mod.state.wait())] = weakref.ref(s_mod)
                    continue
                else:
                    self.logger.fatal(f"{s_mod} cannot be recovered after unexpected termination")
                    # if submodule cannot be recovered it means that application is in inconsistent
                    # and potentially invalid state, terminate it
                    self.terminate_worker()

    def _on_monitor_stop(self, fut: asyncio.Future):
        if fut.cancelled():
            # monitor should never completes, it's run forever until cancelled
            self.logger.debug("Submodules monitor stopped")
            return

        exc = fut.exception()
        if exc:
            self.logger.fatal(f"Submodules monitor terminate with exception: {exc.__class__.__name__}({exc})")
        else:
            self.logger.fatal(f"Submodules monitor terminates unexpectedly")

        self._event_loop.create_task(self.terminate())

    def _setup_logger(self):
        if self.settings.logging:
            logging.config.dictConfig(self.settings.logging)

        now = time()
        self.logger.info(f"Logging configured. Current time is {now} ({datetime.fromtimestamp(now)})")

    def include_router(self, router: TaskRouter):
        """
        Register and setup given processor. Awaits until processor setup is done
        :param router:
        :return:
        """
        self.executor.add_router(router)

    def _setup_signals(self):
        # setup signals using asyncio
        self._event_loop.add_signal_handler(signal.SIGUSR1, partial(self._handle_info_signal, signal.SIGUSR1, None))
        self._event_loop.add_signal_handler(signal.SIGTERM, partial(self._handle_term_signal, signal.SIGTERM, None))
        self._event_loop.add_signal_handler(signal.SIGINT, partial(self._handle_term_signal, signal.SIGTERM, None))

        self._terminate_request = asyncio.Event()

    def get_state(self) -> MutableMapping:
        process = psutil.Process(os.getpid())
        stats = process.as_dict(attrs=(
            'pid', 'status', 'nice', 'io_counters', 'memory_info', 'cpu_times', 'create_time', 'memory_percent',
            'cpu_percent'
        ))

        cpu_limit = process.rlimit(psutil.RLIMIT_CPU)
        mem_limit = process.rlimit(psutil.RLIMIT_RSS)
        threads = threading.enumerate()

        async_tasks = [t for t in asyncio.Task.all_tasks(loop=self._event_loop) if not t.done()]
        process_started = datetime.fromtimestamp(stats['create_time'])

        result = {
            'process': {
                'pid': stats['pid'],
                'status': stats['status'],
                'nice': stats['nice']
            },
            'resources': {
                'io counters': stats['io_counters'],
                'memory usage': naturalsize(stats['memory_info'].rss),
                'memory system [%]': f'{stats["memory_percent"]:.02f} %',
                'cpu times': stats['cpu_times'],
                'cpu usage [%]': f'{stats["cpu_percent"]:.02f} %',
                'started at': str(process_started),
                'uptime': naturaldelta(time() - stats['create_time']),
                'threads': [
                    {'name': t.name, 'daemon': t.daemon, 'alive': t.is_alive()} for t in threads
                ]
            },
            'limits': {},
            'futures': [f"{t._state} {t._coro.cr_code.co_name}() on {t._coro.cr_frame.f_locals['self']}"
                        for t in async_tasks],
            'worker': self.executor.get_state()
        }

        if mem_limit[1] != -1:
            result['limit']['memory'] = {
                'used': f'{(stats["memory_info"].rss/mem_limit[1]):.02f}',
                'limit': f'{naturalsize(mem_limit[0])} | {naturalsize(mem_limit[1])}'
            }

        if cpu_limit[0] != -1:
            result['limit']['cpu'] = {
                'used': f'{(stats["cpu_times"].user/cpu_limit[1])*100:.02f} %',
                'limit': cpu_limit,
            }

        return result

    def _handle_info_signal(self, signum, frame):
        stats = self.get_state()

        sys.stdout.write(f"Executor state information\n"
                         f"--------------------------\n")
        self._print_stats_dict(stats)
        sys.stdout.flush()

    def _print_stats_dict(self, data: Mapping, indentation=1):
        indent = "\t" * indentation
        for name, value in data.items():
            if not value:
                continue

            sys.stdout.write(f'{indent}{name} = ')

            if isinstance(value, dict):
                sys.stdout.write('\n')
                self._print_stats_dict(value, indentation+1)
            elif isinstance(value, (list, tuple)):
                sys.stdout.write('[\n{}]\n'.format(",\n".join(f"\t{v}" for v in value)))
            else:
                sys.stdout.write(f'{value}\n')

    def _handle_term_signal(self, signum, frame):
        self.terminate_worker()

    def terminate_worker(self):
        if self._terminate_request.is_set():
            self.logger.error("Termination signal received again, suicide")

            # restore python sigterm default handler
            signal.signal(signal.SIGTERM, signal.SIG_DFL)

            # send re-send sigterm to terminate process
            os.kill(os.getpid(), signal.SIGTERM)

        self._terminate_request.set()

        self.logger.info(f"Worker termination request received, terminating...")
        stop_job = self._event_loop.create_task(self.terminate())
        stop_job.add_done_callback(self._stop_job_done)

    async def _stop(self):
        # stop submodules monitor
        self.logger.debug("Stopping submodules monitor...")
        self.monitor_task.cancel()

        self.logger.debug("Waiting executor to complete...")
        # wait for executor, because working processors may depends on submodules, so terminate them
        # firstly
        await self.executor.terminate()
        self.logger.info("Executor completed")

        # stop clients, there should be no messages pending on clients as executor is completed
        await self.get_subscriber().terminate()

        if self.settings.publisher is not None:
            await self.get_publisher().terminate()

        stop_jobs = [mod.terminate() for mod in self.modules.values()]
        if stop_jobs:
            self.logger.debug("Waiting for submodules stop jobs to complete...")
            # executor is down, stop all submodules simultaneously, stop jobs should not depends on each other
            await asyncio.wait(stop_jobs)
            self.logger.info("Submodules completed")
        else:
            self.logger.debug("No stop jobs for submodules")

        self.logger.debug("Waiting for manager to stop...")
        await asyncio.wait([asyncio.create_task(self.manager.terminate())])
        self.logger.info("Manager stopped")

        self_task = asyncio.current_task(loop=self._event_loop)
        tasks_left = [t for t in asyncio.all_tasks(self._event_loop) if not t.done() and t != self_task]

        if tasks_left:
            self.logger.error(f"{len(tasks_left)} tasks left in the event loop after stopping all modules")
        else:
            self.logger.debug("No tasks left after stop job, event loop can be stopped clearly")

    def _stop_job_done(self, fut: Future):
        fut.result()
        self._event_loop.stop()

    @classmethod
    def _get_event_loop(cls):
        return asyncio.get_event_loop()



# -*- coding: utf-8 -*-

from datetime import timedelta, datetime
from typing import Union


class TaskControlException(Exception):
    """
    Set of control exceptions raised by processors to notify executor.
    """
    pass


class TaskRetry(TaskControlException):
    """
    Requests task processing retry. Waits optional countdown seconds until next attempt.
    All processed data and stored outputs will be dropped. If retry attempts is reached
    task is backoffed.
    """

    def __init__(self, *, countdown: float = None, progress: Union[float, str, bytes] = None):
        if countdown is not None:
            self.not_before = datetime.utcnow() + timedelta(seconds=countdown)
        else:
            self.not_before = None

        self.progress = progress


class TaskFatal(TaskControlException):
    """
    Processor ends with fatal exception and task is unprocessable now and in the future.
    No retry, output dropped. Log task on backoff queue.
    """

    def __init__(self, *, reason: str = None):
        self.reason = reason


class TaskDrop(TaskControlException):
    """
    Drop this task without logging it on backoff queue.
    """
    pass


class TaskNoHandler(TaskFatal):
    """
    There is no handler to process this task
    """
    pass



# -*- coding: utf-8 -*-
from types import ModuleType
from typing import Optional, Dict, Iterable, Union

from patchwork.node import PatchworkWorker
from patchwork.node.core.exceptions import TaskFatal

try:
    from tortoise import Tortoise, exceptions
except ImportError:
    raise RuntimeError("Tortoise-ORM seems to be not installed")


def register_tortoise(
        worker: PatchworkWorker,
        db_url: str,
        modules: Optional[Dict[str, Iterable[Union[str, ModuleType]]]] = None,
        generate_schemas: bool = False,
        add_exception_handlers: bool = False,
):
    @worker.on_event("startup")
    async def init_orm() -> None:
        await Tortoise.init(db_url=db_url, modules=modules)
        if generate_schemas:
            await Tortoise.generate_schemas()

    @worker.on_event("shutdown")
    async def close_orm() -> None:
        await Tortoise.close_connections()

    if add_exception_handlers:
        @worker.exception_handler(exceptions.OperationalError)
        async def handle_tortoise_orm_exceptions(task, receiver, exc):
            # operational error are related to data, so it make no sense to retry this task
            raise TaskFatal(task=task, reason=f"Tortoise ORM Operational Error: {exc.__class__.__name__}({exc})")



# -*- coding: utf-8 -*-

import asyncio

from aiohttp import web

from .base import Manager


DEFAULT_HOST = "127.0.0.1"
DEFAULT_PORT = 5000


class HttpManager(Manager):
    """
    HTTP manager for the application.
    Supports:
    * prometheus metrics expose
    * cancelling current working tasks
    * config changes though app config provider
    * application termination
    * liveness and rediness probe
    """

    class Config(Manager.Config):
        listen_on: str = f"{DEFAULT_HOST}:{DEFAULT_PORT}"

    server_task: asyncio.Future

    async def _start(self):
        host, port = self.settings.listen_on.split(":")
        if not host:
            host = DEFAULT_HOST

        if not port:
            port = DEFAULT_PORT
        else:
            port = int(port)

        app = web.Application()
        self._setup_http_app(app)
        runner = web.AppRunner(app)
        await runner.setup()
        site = web.TCPSite(runner, host, port)
        await site.start()

        self.logger.info(f"HTTP manager is listening at {host}:{port}")
        self.server_task = self.app_loop.create_task(self._http_server_wait(site, runner))
        self.server_task.add_done_callback(self._on_server_task_done)

    async def _stop(self):
        self.server_task.cancel()
        await self.server_task

    def _on_server_task_done(self, fut):
        exc = fut.exception()
        if exc:
            self.logger.error(f"HTTP manager server terminated with exception {exc.__class__.__name__}({exc})")

    async def _http_server_wait(self, site, runner):
        try:
            await site._server.wait_closed()
        except asyncio.CancelledError:
            await runner.cleanup()

            if not self._is_stopping:
                # server cancelled but component is not stopping? something got wired...
                self.logger.error("HTTP manager server stopped, but manager itself is not stopping")
                # stop the manager
                await self.terminate()

    def _setup_http_app(self, app: web.Application):
        app.router.add_get('/liveness', self._handle_liveness_probe)
        app.router.add_get('/rediness', self._handle_rediness_probe)
        app.router.add_get('/state', self._handle_state)

    async def _handle_liveness_probe(self, request: web.Request) -> web.Response:

        state, details = self.get_health(include_starting=True)
        return web.json_response(details, status=(200 if state else 503))

    async def _handle_rediness_probe(self, request: web.Request) -> web.Response:

        # app is ready if all components are live and not starting
        state, details = self.get_health(include_starting=False)
        resp = web.json_response(details, status=(200 if state else 503))
        return resp

    async def _handle_state(self, request: web.Request) -> web.Response:
        return web.json_response(self.worker.get_state())

    def __repr__(self):
        res = super().__repr__()
        return f"<{res[1:-1]} [{self.settings.listen_on}]>"



# -*- coding: utf-8 -*-
import asyncio

import pytest

from patchwork.node.executor.unit import InlineUnit


@pytest.fixture
async def unit():
    unit = InlineUnit(parent=None)
    await unit.run()
    yield unit
    await unit.terminate()


@pytest.mark.asyncio
async def test_usecase(unit):

    assert unit.is_running
    assert not unit.busy.value

    event = asyncio.Event()
    task_entered = asyncio.Event()

    async def task():
        task_entered.set()
        await event.wait()

    await unit.submit(task())

    # await until processing unit starts task processing
    await task_entered.wait()

    assert unit.busy.value
    event.set()

    # await until processing completes
    await unit.busy.wait_for(False)

    assert not unit.busy.value


@pytest.mark.asyncio
async def test_task_result(unit):
    class Result:
        pass

    async def task(arg):
        return arg

    fut = await unit.submit(task(Result))
    await fut

    assert fut.done()
    assert fut.result() == Result


@pytest.mark.asyncio
async def test_task_exception(unit):

    class CustomException(Exception):
        pass

    async def task():
        raise CustomException()

    fut = await unit.submit(task())

    with pytest.raises(CustomException):
        await fut



# -*- coding: utf-8 -*-

import asyncio
from typing import Mapping

from asyncmock import AsyncMock

from patchwork.core.testutils.mocks import ConfigMock


class ProcessingUnitMock:
    pass


class ExecutorMock:
    pass


class WorkerMock:

    settings: ConfigMock

    def __init__(self, settings: Mapping = None):
        if settings is None:
            settings = {
                'debug': False,
                'processors': []
            }

        self.settings = ConfigMock(settings)

    @property
    def event_loop(self):
        return asyncio.get_event_loop()



# -*- coding: utf-8 -*-

from asyncio import AbstractEventLoop

from patchwork.core.component import Component, OrphanedComponent


class Module(Component):
    """
    Worker level component is named module.
    """

    def __repr__(self):
        return super().__repr__().replace('Component', 'Module')

    @property
    def worker(self):
        """
        Returns worker instance
        :return:
        """
        try:
            return self.parent
        except OrphanedComponent:
            # change exception message
            raise OrphanedComponent("Worker instance gone! This module instance is orphaned and should be destroyed")

    @property
    def app_loop(self) -> AbstractEventLoop:
        """
        Returns worker event loop
        :return:
        """
        return self.worker.event_loop



# -*- coding: utf-8 -*-

import asyncio
import logging
from datetime import datetime
from functools import partial, lru_cache
from itertools import chain
from operator import itemgetter
from typing import MutableMapping, List, Union, Coroutine, Type, Sequence, Dict, Callable, Awaitable, Optional, \
    Iterator, NoReturn

from patchwork.core import Task, AsyncSubscriber
from patchwork.core.config.base import ComponentConfig
from patchwork.core.typing import FuncPath
from patchwork.core.utils import cached_property
from patchwork.node.core import Module
from patchwork.node.core.exceptions import TaskControlException, TaskRetry, TaskFatal, TaskDrop, TaskNoHandler
from patchwork.node.core.router import TaskRouter
from patchwork.node.executor.unit import InlineUnit
from patchwork.node.executor.unit.base import ProcessingUnit

# task finalizer, run when task processing completes, if third argument is None
# means that task has been not processed as there was no processor to handle it
FinalizerType = Callable[[Task, AsyncSubscriber, asyncio.Future], Awaitable]
# middleware is to prepare task processing environment, optional finalizer is a method which
# will be called to finalize task processing. Middleware may return new finalizer which should
# invoke given finalizer internally for proper task finalization
# if middleware does not return new finalizer, passed finalizer will be used
# middlewares are run on the worker async loop, before passing task to processing unit!
# keep this in mind as middleware execution may slow down whole executor
MiddlewareType = Callable[[Task, AsyncSubscriber, FinalizerType], Awaitable[Optional[FinalizerType]]]

ExceptionHandler = Callable[[Task, Exception], Awaitable]


async def check_task_middleware(task: Task, receiver: AsyncSubscriber, finalizer: FinalizerType):
    now = datetime.utcnow().timestamp()
    if task.meta.HasField('expires') and now > task.meta.expires.ToSeconds():
        # task has expired
        raise TaskFatal(reason="Task has expired")
    elif task.meta.HasField('not_before') and now < task.meta.not_before.ToSeconds():
        # task is not ready yet
        raise TaskRetry(countdown=(now - task.meta.not_before))


async def default_task_exception_handler(task: Task, exception: Exception) -> NoReturn:
    if isinstance(exception, TaskControlException):
        raise exception

    # request task retry
    raise TaskRetry()


class Idempotent:

    idempotent = True

    def __init__(self, fn):
        self._fn = fn

    def __call__(self, task: Task, exc: Exception):
        try:
            return self._fn(task, exc)
        except TaskControlException:
            raise RuntimeError(f"Idempotent exception handlers can't raise TaskControlExceptions")

    def __repr__(self):
        return f"idempotent<{repr(self._fn)}>"


class BaseExecutor(Module):
    """
    Executor is responsible of taking tasks from supported queue, spawn and manage workers
    to run these tasks and implement ability to send new tasks to the queue though clients.
    """

    class Config(Module.Config):

        # executor termination timeout in seconds
        terminate_timeout: int = 10
        unit: ComponentConfig[Type[ProcessingUnit]] = ComponentConfig(InlineUnit)
        middlewares: Sequence[FuncPath] = [
            FuncPath(check_task_middleware)
        ]

    logger_name = 'executor'
    _loop_task: asyncio.Future
    _routers: List[TaskRouter]
    _unit_monitor_task: asyncio.Future

    def __init__(self, *, parent, **options):
        super().__init__(parent=parent, **options)

        self._terminate_request = asyncio.Event()
        self._current_tasks = set()
        self._routers = []
        self._middlewares: List[MiddlewareType] = []
        self._exc_handlers: Dict[type, List] = {}

        for middleware in self.settings.middlewares:
            self.add_middleware(middleware)

        self.add_exception_handler(Exception, default_task_exception_handler)

    def __repr__(self):
        res = super().__repr__()
        return f"<{res[1:-1]} " \
               f"unit={self.unit.__class__.__name__}]>"

    def get_state(self) -> MutableMapping:
        return {
            'worker': self.unit.get_state(),
            'current_tasks': len(self._current_tasks)
        }

    @cached_property
    def unit(self) -> ProcessingUnit:
        return self.settings.unit.instantiate(parent=self)

    def add_router(self, router: TaskRouter):
        router.bind(self.worker)
        self._routers.append(router)

    def add_middleware(self, middleware: MiddlewareType):
        self._middlewares.append(middleware)

    def add_exception_handler(self, exception: type, handler: ExceptionHandler, idempotent: bool = False):
        """
        Register a handler for given exception type (and subtypes). Exception handlers are called
        in LIFO manner - last registered handler is firstly called.

        If any handler raises one of TaskControlException it's handled by the executor and then no more exception
        handlers are executed (except idempotent ones)
        :param exception:
        :param handler:
        :param idempotent: if True, handler will be executed even if TaskControlException has been raised by one
        of previously executed handlers.
            :: warning
                Idempotent handlers CAN'T raise TaskControlExceptions.
        :return:
        """
        if exception not in self._exc_handlers:
            self._exc_handlers[exception] = []

        if idempotent:
            handler = Idempotent(handler)
        self._exc_handlers[exception].insert(0, handler)

    async def _start(self):
        """
        Setups executor
        :return:
        """

        # clear terminate request flag
        self._terminate_request.clear()

        # start processing unit
        await self.unit.run()

        self._unit_monitor_task = asyncio.create_task(self._unit_monitor())

        # put executor loop task on event loop
        self._loop_task = asyncio.ensure_future(self._loop(), loop=self.app_loop)
        self._loop_task.add_done_callback(self._loop_completed_callback)

    async def _unit_monitor(self):
        try:
            await self.unit.state.wait_for(False)
        except asyncio.CancelledError:
            return

        self.logger.error(f"Executor processing unit unexpectedly terminated. Executor can't continue")

        self.harakiri()

    def _loop_completed_callback(self, fut: asyncio.Future):

        if fut.cancelled():
            # loop should never be cancelled
            raise AssertionError("DO NOT CANCEL executor loop! Set termination_request event instead to make"
                                 "sure that loop is stopped cleanly and no message was lost")

        exc = fut.exception()
        if exc:
            self.logger.error(f"Executor loop completed with exception: {exc.__class__.__name__}({exc})")
            self.harakiri()

    async def _stop(self):
        """
        Stops executor and all running tasks, when method returns event loop can be stopped
        :return:
        """

        # set terminate request flag
        self._terminate_request.set()
        self.logger.debug(f"Executor termination requested, "
                          f"waiting {self.settings.terminate_timeout}s to complete...")

        # wait for termination
        done, pending = await asyncio.wait([self._loop_task], timeout=self.settings.terminate_timeout)

        if pending:
            self.logger.warning("Loop task termination timeout, cancelling")

            # this raises CancelledError in loop task, there is no way to predict where exactly in code
            # it happens
            self._loop_task.cancel()

            # timeout to be safe
            done, pending = await asyncio.wait([self._loop_task], timeout=self.settings.terminate_timeout)

            if pending:
                self.logger.error("Unable to terminate executor main loop")

        # stop processing unit, executor loop is stopped so no more task may come,
        # drop what unit has queued and what is currently processing
        # there should be no more job to do about tasks, commit() method has been not
        # called so tasks assigned to the processing unit should be not lost
        try:
            await asyncio.wait_for(self.unit.busy.wait_for(False), timeout=self.settings.terminate_timeout)
        except asyncio.TimeoutError:
            self.logger.warning(f"Processing unit is still busy, timeout elapsed")

        self._unit_monitor_task.cancel()
        await self.unit.terminate()

        # wait for tasks callbacks to complete (commit, retry..., the last job is removing task from the
        # _current_tasks set)
        if self._current_tasks:
            await asyncio.wait(self._current_tasks)

    async def _loop(self):
        """
        This is an executor main loop task. Waits on all listen_on subscribers simultaneously for
        incoming task.
        DO NOT CANCEL THIS CORO as it can break flow in a way which may leads to message lost.
        Set self._termination_request event to request the loop termination.
        :return:
        """
        # future of termination request
        term_future = asyncio.create_task(self._terminate_request.wait())
        subscriber = self.worker.get_subscriber()

        while True:

            try:
                done, pending = await asyncio.wait((asyncio.create_task(subscriber.get()), term_future,),
                                                   return_when=asyncio.FIRST_COMPLETED)
            except Exception as e:
                self.logger.error(f"Client get() raises exception: {e.__class__.__name__}({e})",
                                  exc_info=True)
            else:
                for task in done:
                    if task == term_future:
                        # termination future done, even if there is more done futures skip them
                        # worker is going down
                        for fut in pending:
                            fut.cancel()
                        return

                    await self._handle_done_waiter(task, subscriber)

    async def _handle_done_waiter(self, future: asyncio.Future, receiver: AsyncSubscriber) -> bool:
        """
        Handles done waiter future of the client
        :param future:
        :param receiver:
        :return: True if loop should still listen on this client, False otherwise
        """

        if future.cancelled():
            # ygm? maybe something else disabled this client... do not put it back
            self.logger.info(f"Waiter of client {receiver} cancelled")
            await receiver.terminate()
            return False

        exc = future.exception()
        if exc:
            self.logger.error(f"Client {receiver} waiter completed with exception: "
                              f"{exc.__class__.__name__}({exc})")
            return True

        task: Task = future.result()
        self.logger.debug(f"Client {receiver} waiter returned a task message: {task.uuid}")

        try:
            await self.handle(task, receiver)
        except Exception as e:
            # log error and rollback task if something went wrong during internal scheduling
            self.logger.error(f"Unable to handle task '{task.uuid}', exception raised: {e.__class__.__name__}({e})")
            try:
                await self.retry(task)
                # TODO: commit or not?
            except Exception as e:
                self.logger.error(f"Unable to retry unhandled task '{task}', "
                                  f"exception raised: {e.__class__.__name__}({e})")
                await self.backoff_task(task, reason="Internal retry failed")
                # TODO: commit or not?

        return True

    async def retry(self, task: Task, *, countdown: float = None,
                    not_before: datetime = None, bump_attempts: bool = True) -> Union[Task, None]:
        """
        Retries given task by cloning it and sending through publisher.
        :param task:
        :param countdown:
        :param not_before:
        :param bump_attempts:
        :return:
        """
        new_task = Task()
        new_task.CopyFrom(task)
        new_task.uuid = ''

        if bump_attempts:
            new_task.meta.attempt += 1
            if task.meta.max_retries > 0 and new_task.meta.attempt > new_task.meta.max_retries:
                await self.backoff_task(task, reason="Max number of retries exceeded")
                return

        if countdown is not None:
            assert not_before is None, "both `countdown` and `not_before` arguments cannot be set at the same time"

            not_before = datetime.utcnow().timestamp() + countdown

        if not_before is not None:
            new_task.meta.not_before.FromDatetime(not_before)

        publisher = self.worker.get_publisher()
        await publisher.send_task(new_task)
        return new_task

    async def backoff_task(self, task, reason: str = None):
        """
        Stores given task on backoff queue. If task cannot be executed for some reason (eg retries limit exceeded)
        can be logged and saved in special place for future investigation. This method allows to notify administrator
        that something has been not executed and should be manually verified.

        In base implementation this method just logs task using 'backoff' standard python logger and passes
        task in `extra` argument, which means that in basic configuration task contents might be lost.
        :param task:
        :param reason:
        :return:
        """
        backoff_log = logging.getLogger('backoff')
        backoff_log.log(logging.NOTSET,
                        f"Task '{task.uuid}' backoffed.{(' Reason: ' + reason) if reason is not None else ''}",
                        extra={'task': task})
        self.logger.debug(f"Task '{task.uuid}' backoffed.{(' Reason: ' + reason) if reason is not None else ''}")

    async def handle(self, task: Task, receiver: AsyncSubscriber) -> Union[asyncio.Future, None]:
        """
        Executes given task, handles all exceptions which may rise during task
        processing.

        :: warning:
            This method may raise exception if task scheduling fail.

        :: note:
            This method awaits only for task scheduling, not task execution itself. When returns task
            is internally scheduled on processing unit. If you need task execution result await on returned
            future.

        :param task:
        :param receiver:
        :return: future of task processing function or None if processing was not scheduled because of
                 task control exception
        """
        finalizer = self._handle_task_result
        task_future = self.app_loop.create_future()

        try:
            for middleware in self._middlewares:
                finalizer = (await middleware(task, receiver, finalizer)) or finalizer

            finalizer = partial(finalizer, task=task, receiver=receiver)

            exec_fut = await self._execute(task, receiver)
        except Exception as exc:
            exec_fut = self.app_loop.create_future()
            exec_fut.set_exception(exc)

        exec_fut.add_done_callback(
            partial(self._task_done_callback, finalizer=finalizer(processing_fut=exec_fut), task_future=task_future)
        )

        self._current_tasks.add(task_future)
        task_future.add_done_callback(lambda f: self._current_tasks.remove(task_future))

        return task_future

    async def _execute(self, task: Task, receiver: AsyncSubscriber):
        """
        Executes a task on given processor
        :param task:
        :param receiver:
        :return:
        """
        for router in self._routers:
            processor = router.handle(task, receiver)
            if processor is not None:
                # await for submit which may hold if processing unit is full
                pu_fut = await self.unit.submit(processor)
                return pu_fut

        self.logger.error(f"No processor to handle message type '{task.task_type}'")
        raise TaskNoHandler()

    def _task_done_callback(self,
                            processing_fut: asyncio.Future,
                            finalizer: Coroutine,
                            task_future: asyncio.Future
                            ):
        """
        Handles value returned by task processing future.
        :param finalizer:
        :param processing_fut:
        :param task_future:
        :return:
        """
        def _callback(f: asyncio.Future):
            if f.cancelled():
                # if finalizer has been cancelled to not pass result from task
                # processing future to task future, set it to None as there is no ready result as finalizers
                # were not completed
                task_future.set_result(None)
                return

            exc = f.exception()
            if exc is not None:
                # if finalizer ends with unhandled exception it's internal fatal error
                # kill the executor
                task_future.set_result(None)
                self.logger.fatal(f"Can't handle task result: {exc.__class__.__name__}({exc})")
                self.harakiri(exc)

            # otherwise pass processing future result to task result
            if processing_fut.cancelled():
                task_future.cancel()
            elif processing_fut.exception():
                task_future.set_exception(processing_fut.exception())
            else:
                task_future.set_result(processing_fut.result())

        handle_fut = self.app_loop.create_task(finalizer)
        handle_fut.add_done_callback(_callback)

    @lru_cache(maxsize=None)
    def _get_exception_handler(self, exception) -> Iterator[ExceptionHandler]:
        handlers = []

        for exc_type, callbacks in self._exc_handlers.items():
            if not isinstance(exception, exc_type):
                continue

            done = False
            for idx, item in enumerate(handlers):
                if exc_type == item[0]:
                    item[1].extend(callbacks)
                    done = True
                    break

                if issubclass(exc_type, item[0]):
                    handlers.insert(idx, (exc_type, list(callbacks)))
                    done = True
                    break

            if not done:
                handlers.append((exc_type, list(callbacks)))

        return chain.from_iterable(map(itemgetter(1), handlers))

    async def _handle_task_result(self, task: Task, receiver: AsyncSubscriber, processing_fut: asyncio.Future):
        """
        Handles task future result.
        :param task:
        :param receiver:
        :param processing_fut:
        :return:
        """
        exc = processing_fut.exception()
        if processing_fut.cancelled():
            self.logger.debug(f"Task '{task}' execution cancelled internally")
            return
        elif exc is not None:
            self.logger.warning(f"Task '{task}' execution failed with exception {exc.__class__.__name__}({exc})")
            handlers = self._get_exception_handler(exc)
            finalized = False

            for handler in handlers:
                if finalized and not getattr(handler, 'idempotent', False):
                    # if task is finalized (there was a TCE raised and handled), run only
                    # handlers declared as idempotent
                    continue

                try:
                    await handler(task, exc)
                except TaskControlException as control_exc:
                    await self._handle_task_control(task, control_exc)
                    finalized = True
                except Exception as exc:
                    self.logger.exception(f"exception handler failed: {exc.__class__.__name__}({exc})")
                    if self.worker.debug:
                        raise exc

        await self._commit_task(task, receiver)

    async def _handle_task_control(self, task: Task, exc: Exception):
        """
        Takes appropriate action for the task depending on given task control exception.
        This method commits task.
        :param task:
        :param exc:
        :return:
        """
        if isinstance(exc, TaskRetry):
            # schedule new task
            await self.retry(task, not_before=exc.not_before)
        elif isinstance(exc, TaskFatal):
            await self.backoff_task(task, reason=(exc.reason or "Task processing fatal error"))
            # backoff just logs task, commit it to truly remove from the queue
        elif isinstance(exc, TaskDrop):
            pass
        elif isinstance(exc, TaskControlException):
            self.logger.error(f"Unsupported task control exception '{exc.__class__.__name__}({exc})'")
            if self.worker.debug:
                raise ValueError("unknown task control exception")

    async def _commit_task(self, task: Task, receiver: AsyncSubscriber):
        """
        Commits the task on the underlying queue.
        :param task:
        :return:
        """
        await receiver.commit(task)

    def harakiri(self, cause: BaseException = None):
        """
        Stops executor immediately due to critical internal error.
        Call this method if something went so wrong that there is no better way than killing this executor
        which cause clients disconnection from message broker and leaving uncommitted all tasks which
        are in progress or pending. Broker should consider this worker as dead and stop delivering new messages.
        :return:
        """
        if self.worker.debug and cause:
            # in debug mode raise SystemExit to blow up
            raise SystemExit() from cause

        self.logger.error("Executor harakiri requested")
        self.app_loop.create_task(self.terminate())



import os

from pip._internal.req import parse_requirements
from setuptools import setup, find_namespace_packages

version = os.environ.get('CI_COMMIT_TAG', f"0.0.dev{os.environ['CI_JOB_ID']}")
if '-' in version:
    # version tag should be like: node-1.0.0
    version = version.split('-')[1]

last_part = version.split('.')[-1]
version_classifier = "5 - Production/Stable"

if 'dev' in last_part:
    version_classifier = "2 - Pre-Alpha"
elif 'a' in last_part:
    version_classifier = "3 - Alpha"
elif 'b' in last_part:
    version_classifier = "4 - Beta"

with open("README.md", "r") as fh:
    long_description = fh.read()

dependencies = parse_requirements('requirements.txt', session=None)

setup(
    name='patchwork-node',
    version=version,
    packages=['patchwork.node'] + find_namespace_packages(include="patchwork.node.*"),
    url='',
    author='Pawel Pecio',
    author_email='',
    long_description=long_description,
    long_description_content_type="text/markdown",
    description='Worker node for Patchwork - The distributed asynchronous microframework',
    zip_safe=False,
    install_requires=[str(ir.requirement) for ir in dependencies],
    classifiers=[
        f"Development Status :: {version_classifier}",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Software Development :: Libraries :: Application Frameworks"
    ]
)



# -*- coding: utf-8 -*-
import asyncio
from unittest import mock

import pytest
from patchwork.core import Task, AsyncSubscriber

from patchwork.node.core.exceptions import TaskRetry, TaskDrop, TaskFatal, TaskControlException
from patchwork.node.executor import BaseExecutor
from tests.router import test_router, CustomException, UnknownTCE


class WorkerMock:

    debug = True

    def __init__(self):
        self.get_publisher = mock.Mock(return_value=mock.AsyncMock())

    @property
    def event_loop(self):
        return asyncio.get_running_loop()


@pytest.fixture
async def worker_mock():
    return WorkerMock()


@pytest.fixture
async def executor(worker_mock):
    executor = BaseExecutor(parent=worker_mock)
    executor.add_router(test_router)
    await executor.run()
    yield executor
    await executor.terminate()


@pytest.fixture
def subscriber():
    return mock.AsyncMock(spec=AsyncSubscriber)


@pytest.mark.asyncio
async def test_task_success(executor, subscriber):
    task = Task(task_type='noop')
    fut = await executor.handle(task, subscriber)
    await fut
    subscriber.commit.assert_called_once_with(task)


@pytest.mark.asyncio
async def test_task_processor_error(executor, subscriber):
    executor.retry = mock.AsyncMock()

    task = Task(task_type='exception')
    fut = await executor.handle(task, subscriber)

    with pytest.raises(CustomException):
        await fut

    subscriber.commit.assert_called_once_with(task)
    executor.retry.assert_called_once()


@pytest.mark.asyncio
async def test_task_retry(executor, subscriber):

    task = Task(task_type='retry')
    fut = await executor.handle(task, subscriber)

    with pytest.raises(TaskRetry):
        # exception from task processor should be passed to task future
        await fut

    subscriber.commit.assert_called_once_with(task)
    executor.worker.get_publisher.assert_called_once()


@pytest.mark.asyncio
async def test_task_drop(executor, subscriber):

    executor.retry = mock.AsyncMock()
    executor.backoff_task = mock.AsyncMock()

    task = Task(task_type='drop')
    fut = await executor.handle(task, subscriber)

    with pytest.raises(TaskDrop):
        # exception from task processor should be passed to task future
        await fut

    subscriber.commit.assert_called_once_with(task)
    assert not executor.worker.get_publisher.called
    assert not executor.retry.called
    assert not executor.backoff_task.called


@pytest.mark.asyncio
async def test_task_fatal(executor, subscriber):

    executor.retry = mock.AsyncMock()
    executor.backoff_task = mock.AsyncMock()

    task = Task(task_type='fatal')
    fut = await executor.handle(task, subscriber)

    with pytest.raises(TaskFatal):
        # exception from task processor should be passed to task future
        await fut

    subscriber.commit.assert_called_once_with(task)
    executor.backoff_task.assert_called_with(task, reason='Task processing fatal error')

    assert not executor.worker.get_publisher.called
    assert not executor.retry.called


@pytest.mark.asyncio
async def test_task_unknown_task_control_exc(executor, subscriber):

    executor.harakiri = mock.Mock()
    executor.retry = mock.AsyncMock()

    task = Task(task_type='unknown_tce')
    fut = await executor.handle(task, subscriber)

    await fut

    assert not subscriber.commit.called, \
        "on executor internal error no commit should be done"

    assert not executor.retry.called, \
        "on executor internal error no retry should be done"

    executor.harakiri.assert_called_once()


@pytest.mark.asyncio
async def test_task_exc_handler_retry(executor, subscriber):
    async def custom_exc_handler(*args):
        raise TaskRetry()

    executor.retry = mock.AsyncMock()

    task = Task(task_type='exception')
    executor.add_exception_handler(CustomException, custom_exc_handler)

    fut = await executor.handle(task, subscriber)
    with pytest.raises(CustomException):
        await fut

    executor.retry.assert_called_with(task, not_before=None)
    subscriber.commit.assert_called_once_with(task)


@pytest.mark.asyncio
async def test_task_exc_handler_fatal(executor, subscriber):
    async def custom_exc_handler(*args):
        raise TaskFatal()

    executor.backoff_task = mock.AsyncMock()

    task = Task(task_type='exception')
    executor.add_exception_handler(CustomException, custom_exc_handler)

    fut = await executor.handle(task, subscriber)
    with pytest.raises(CustomException):
        await fut

    executor.backoff_task.assert_called_with(task, reason='Task processing fatal error')
    subscriber.commit.assert_called_once_with(task)


@pytest.mark.asyncio
async def test_task_exc_handler_drop(executor, subscriber):
    async def custom_exc_handler(*args):
        raise TaskDrop()

    task = Task(task_type='exception')
    executor.add_exception_handler(CustomException, custom_exc_handler)

    fut = await executor.handle(task, subscriber)
    with pytest.raises(CustomException):
        await fut

    subscriber.commit.assert_called_once_with(task)


@pytest.mark.asyncio
async def test_task_exc_handler_unknown_control_exc(executor, subscriber):
    async def custom_exc_handler(*args):
        raise UnknownTCE()

    executor.harakiri = mock.Mock()
    task = Task(task_type='exception')
    executor.add_exception_handler(CustomException, custom_exc_handler)

    fut = await executor.handle(task, subscriber)
    await fut

    executor.harakiri.assert_called_once()


@pytest.mark.asyncio
async def test_task_exc_handler_error(executor, subscriber):
    async def custom_exc_handler(*args):
        raise Exception()

    executor.harakiri = mock.Mock()

    task = Task(task_type='exception')
    executor.add_exception_handler(CustomException, custom_exc_handler)

    fut = await executor.handle(task, subscriber)
    await fut

    executor.harakiri.assert_called_once()


@pytest.mark.asyncio
async def test_task_exc_handler_success(executor, subscriber):
    async def custom_exc_handler(*args):
        pass

    task = Task(task_type='exception')
    executor.add_exception_handler(CustomException, custom_exc_handler)

    fut = await executor.handle(task, subscriber)
    with pytest.raises(CustomException):
        await fut

    subscriber.commit.assert_called_once_with(task)



# -*- coding: utf-8 -*-

from typing import Tuple, Mapping

from patchwork.node.core import Module


class Manager(Module):
    """
    Worker manager to manage the worker remotely
    """

    async def _start(self):
        pass

    async def _stop(self):
        pass

    def get_health(self, include_starting: bool = True) -> Tuple[bool, Mapping[str, bool]]:
        """
        Return health of all worker components.
        :param include_starting: If true starting components are also considered as health
        :return:
        """

        state = {}

        from patchwork.node import PatchworkWorker
        worker: PatchworkWorker = self.worker

        state['worker'] = worker.is_running or (worker.is_starting and include_starting)

        # until worker is running, monitor_task may not be created
        state['monitor'] = not worker.monitor_task.done() if worker.is_running else True

        for mod_name, mod_instance in worker.settings.modules.items():
            state[mod_name] = mod_instance.is_running or (mod_instance.is_starting and include_starting)

        return all(state.values()), state
